import copy
import pickle
import random

from collections import OrderedDict
from unittest.mock import PropertyMock, patch

import pytest

from faker import Faker
from faker.config import DEFAULT_LOCALE
from faker.generator import Generator


class TestFakerProxyClass:
    """Test Faker proxy class"""

    def test_unspecified_locale(self):
        fake = Faker()
        assert len(fake.locales) == 1
        assert len(fake.factories) == 1
        assert fake.locales[0] == DEFAULT_LOCALE

    def test_locale_as_string(self):
        locale = "en_US"
        fake = Faker(locale)
        assert len(fake.locales) == 1
        assert len(fake.factories) == 1
        assert fake.locales[0] == locale

    def test_locale_as_list(self):
        locale = ["en-US", "en_PH", "ja_JP", "de-DE"]
        expected = ["en_US", "en_PH", "ja_JP", "de_DE"]
        fake = Faker(locale)
        assert fake.locales == expected
        assert len(fake.factories) == len(expected)

        locale = ["en-US", "en_PH", "ja_JP", "de-DE", "ja-JP", "de_DE", "en-US"] * 3
        expected = ["en_US", "en_PH", "ja_JP", "de_DE"]
        fake = Faker(locale)
        assert fake.locales == expected
        assert len(fake.factories) == len(expected)

    def test_locale_as_list_invalid_value_type(self):
        locale = [1, 2]
        with pytest.raises(TypeError) as exc:
            Faker(locale)
        assert str(exc.value) == 'The locale "1" must be a string.'

    def test_locale_as_ordereddict(self):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )

        fake = Faker(locale)
        assert fake.locales == ["de_DE", "en_US", "en_PH", "ja_JP"]
        assert len(fake.factories) == 4
        assert fake.weights == [3, 2, 1, 5]

        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
                ("de-DE", 4),
                ("ja-JP", 2),
                ("en-US", 1),
            ]
        )
        fake = Faker(locale)
        assert fake.locales == ["de_DE", "en_US", "en_PH", "ja_JP"]
        assert len(fake.factories) == 4
        assert fake.weights == [4, 1, 1, 2]

    def test_invalid_locale(self):
        with pytest.raises(AttributeError):
            Faker("foo_Bar")

        with pytest.raises(AttributeError):
            Faker(["en_US", "foo_Bar"])

        with pytest.raises(AttributeError):
            Faker(
                OrderedDict(
                    [
                        ("de_DE", 3),
                        ("en-US", 2),
                        ("en-PH", 1),
                        ("foo_Bar", 5),
                    ]
                )
            )

    def test_items(self):
        locale = ["de_DE", "en-US", "en-PH", "ja_JP", "de-DE", "ja-JP", "en-US"]
        processed_locale = list({code.replace("-", "_") for code in locale})
        fake = Faker(locale)
        for locale_name, factory in fake.items():
            assert locale_name in processed_locale
            assert isinstance(factory, (Generator, Faker))

    def test_dunder_getitem(self):
        locale = ["de_DE", "en-US", "en-PH", "ja_JP"]
        fake = Faker(locale)

        for code in locale:
            assert isinstance(fake[code], (Generator, Faker))

        with pytest.raises(KeyError):
            fake["en_GB"]

    def test_seed_classmethod(self):
        fake = Faker()

        # Verify `seed()` is not callable from a class instance
        with pytest.raises(TypeError):
            fake.seed(0)

        # Verify calls to `seed()` from a class object are proxied properly
        with patch("faker.generator.Generator.seed") as mock_seed:
            mock_seed.assert_not_called()
            Faker.seed(0)
            mock_seed.assert_called_once_with(0)

    def test_seed_class_locales(self):
        Faker.seed(2043)
        count = 5
        fake = Faker(["en_GB", "fr_FR", "en_IN"])
        first_list = [fake.name() for _ in range(count)]
        # We convert the list to a set to remove duplicates and ensure
        # that we have exactly `count` unique fake values
        assert len(set(first_list)) == count

        Faker.seed(2043)
        fake = Faker(["en_GB", "fr_FR", "en_IN"])
        second_list = [fake.name() for _ in range(count)]

        assert first_list == second_list

    def test_seed_instance(self):
        locale = ["de_DE", "en-US", "en-PH", "ja_JP"]
        fake = Faker(locale)

        with patch("faker.generator.Generator.seed_instance") as mock_seed_instance:
            mock_seed_instance.assert_not_called()
            fake.seed_instance(0)

            # Verify `seed_instance(0)` was called 4 times (one for each locale)
            calls = mock_seed_instance.call_args_list
            assert len(calls) == 4
            for call in calls:
                args, kwargs = call
                assert args == (0,)
                assert kwargs == {}

    def test_seed_locale(self):
        from faker.generator import random as shared_random_instance

        locale = ["de_DE", "en-US", "en-PH", "ja_JP"]
        fake = Faker(locale)

        # Get current state of each factory's random instance
        states = {}
        for locale, factory in fake.items():
            states[locale] = factory.random.getstate()

        # Create a new random instance for en_US factory with seed value
        fake.seed_locale("en_US", 0)

        for locale, factory in fake.items():
            # en_US factory should have changed
            if locale == "en_US":
                assert factory.random != shared_random_instance
                assert factory.random.getstate() != states[locale]

            # There should be no changes for the rest
            else:
                assert factory.random == shared_random_instance
                assert factory.random.getstate() == states[locale]

    def test_single_locale_proxy_behavior(self):
        fake = Faker()
        internal_factory = fake.factories[0]

        # Test if `Generator` attributes are proxied properly
        for attr in fake.generator_attrs:
            assert getattr(fake, attr) == getattr(internal_factory, attr)

        # Test if `random` getter and setter are proxied properly
        tmp_random = fake.random
        assert internal_factory.random != 1
        fake.random = 1
        assert internal_factory.random == 1
        fake.random = tmp_random

        # Test if a valid provider method is proxied properly
        # Factory selection logic should not be triggered
        with patch("faker.proxy.Faker._select_factory") as mock_select_factory:
            mock_select_factory.assert_not_called()
            assert fake.name == internal_factory.name
            fake.name()
            mock_select_factory.assert_not_called()

    def test_multiple_locale_proxy_behavior(self):
        fake = Faker(["de-DE", "en-US", "en-PH", "ja-JP"])

        # `Generator` attributes are not implemented
        for attr in fake.generator_attrs:
            with pytest.raises(NotImplementedError):
                getattr(fake, attr)

        # The `random` getter is not implemented
        with pytest.raises(NotImplementedError):
            random = fake.random
            random.seed(0)

        # The `random` setter is not implemented
        with pytest.raises(NotImplementedError):
            fake.random = 1

    def test_multiple_locale_caching_behavior(self):
        fake = Faker(["de_DE", "en-US", "en-PH", "ja_JP"])

        with patch("faker.proxy.Faker._map_provider_method", wraps=fake._map_provider_method) as mock_map_method:
            mock_map_method.assert_not_called()
            assert not hasattr(fake, "_cached_name_mapping")

            # Test cache creation
            fake.name()
            assert hasattr(fake, "_cached_name_mapping")
            mock_map_method.assert_called_once_with("name")

            # Test subsequent cache access
            with patch.object(Faker, "_cached_name_mapping", create=True, new_callable=PropertyMock) as mock_cached_map:
                # Keep test fast by patching the cached mapping to return something simpler
                mock_cached_map.return_value = [fake["en_US"]], [1]
                for _ in range(100):
                    fake.name()

                # Python's hasattr() internally calls getattr()
                # So each call to name() accesses the cached mapping twice
                assert mock_cached_map.call_count == 200

    @patch("faker.proxy.Faker._select_factory_choice")
    @patch("faker.proxy.Faker._select_factory_distribution")
    def test_multiple_locale_factory_selection_no_weights(self, mock_factory_distribution, mock_factory_choice):
        fake = Faker(["de_DE", "en-US", "en-PH", "ja_JP"])

        # There are no distribution weights, so factory selection logic will use `random.choice`
        # if multiple factories have the specified provider method
        with patch("faker.proxy.Faker._select_factory", wraps=fake._select_factory) as mock_select_factory:
            mock_select_factory.assert_not_called()
            mock_factory_distribution.assert_not_called()
            mock_factory_choice.assert_not_called()

            # All factories for the listed locales have the `name` provider method
            fake.name()
            mock_select_factory.assert_called_once_with("name")
            mock_factory_distribution.assert_not_called()
            mock_factory_choice.assert_called_once_with(fake.factories)
            mock_select_factory.reset_mock()
            mock_factory_distribution.reset_mock()
            mock_factory_choice.reset_mock()

            # Only `en_PH` factory has provider method `luzon_province`, so there is no
            # need for `random.choice` factory selection logic to run
            fake.luzon_province()
            mock_select_factory.assert_called_with("luzon_province")
            mock_factory_distribution.assert_not_called()
            mock_factory_choice.assert_not_called()
            mock_select_factory.reset_mock()
            mock_factory_distribution.reset_mock()
            mock_factory_choice.reset_mock()

            # Both `en_US` and `ja_JP` factories have provider method `zipcode`
            fake.zipcode()
            mock_select_factory.assert_called_once_with("zipcode")
            mock_factory_distribution.assert_not_called()
            mock_factory_choice.assert_called_once_with(
                [fake["en_US"], fake["ja_JP"]],
            )

    @patch("faker.proxy.Faker._select_factory_choice")
    @patch("faker.proxy.Faker._select_factory_distribution")
    def test_multiple_locale_factory_selection_with_weights(self, mock_factory_distribution, mock_factory_choice):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )
        fake = Faker(locale)
        mock_factory_distribution.assert_not_called()
        mock_factory_choice.assert_not_called()

        # Distribution weights have been specified, so factory selection logic will use
        # `choices_distribution` if multiple factories have the specified provider method
        with patch("faker.proxy.Faker._select_factory", wraps=fake._select_factory) as mock_select_factory:
            # All factories for the listed locales have the `name` provider method
            fake.name()
            mock_select_factory.assert_called_once_with("name")
            mock_factory_distribution.assert_called_once_with(fake.factories, fake.weights)
            mock_factory_choice.assert_not_called()

    @patch("faker.proxy.Faker._select_factory_choice")
    @patch("faker.proxy.Faker._select_factory_distribution")
    def test_multiple_locale_factory_selection_single_provider(self, mock_factory_distribution, mock_factory_choice):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )
        fake = Faker(locale)

        # Distribution weights have been specified, so factory selection logic will use
        # `choices_distribution` if multiple factories have the specified provider method
        with patch("faker.proxy.Faker._select_factory", wraps=fake._select_factory) as mock_select_factory:
            # Only `en_PH` factory has provider method `luzon_province`, so there is no
            # need for `choices_distribution` factory selection logic to run
            fake.luzon_province()
            mock_select_factory.assert_called_once_with("luzon_province")
            mock_factory_distribution.assert_not_called()
            mock_factory_choice.assert_not_called()

    @patch("faker.proxy.Faker._select_factory_choice")
    @patch("faker.proxy.Faker._select_factory_distribution")
    def test_multiple_locale_factory_selection_shared_providers(self, mock_factory_distribution, mock_factory_choice):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )
        fake = Faker(locale)

        with patch("faker.proxy.Faker._select_factory", wraps=fake._select_factory) as mock_select_factory:
            # Both `en_US` and `ja_JP` factories have provider method `zipcode`
            fake.zipcode()
            mock_select_factory.assert_called_once_with("zipcode")
            mock_factory_distribution.assert_called_once_with([fake["en_US"], fake["ja_JP"]], [2, 5])
            mock_factory_choice.assert_not_called()

    def test_multiple_locale_factory_selection_unsupported_method(self):
        fake = Faker(["en_US", "en_PH"])
        with pytest.raises(AttributeError):
            fake.obviously_invalid_provider_method_a23f()

    @patch("random.Random.choice")
    @patch("random.Random.choices")
    def test_weighting_disabled_single_choice(self, mock_choices_fn, mock_choice_fn):
        fake = Faker(use_weighting=False)
        fake.first_name()
        mock_choice_fn.assert_called()
        mock_choices_fn.assert_not_called()

    @patch("random.Random.choice")
    @patch("random.Random.choices", wraps=random.Random().choices)
    def test_weighting_disabled_with_locales(self, mock_choices_fn, mock_choice_fn):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )
        fake = Faker(locale, use_weighting=False)
        fake.first_name()
        mock_choices_fn.assert_called()  # select provider
        mock_choice_fn.assert_called()  # select within provider

    @patch("random.Random.choice")
    @patch("random.Random.choices", wraps=random.Random().choices)
    def test_weighting_disabled_multiple_locales(self, mock_choices_fn, mock_choice_fn):
        locale = OrderedDict(
            [
                ("de_DE", 3),
                ("en-US", 2),
                ("en-PH", 1),
                ("ja_JP", 5),
            ]
        )
        fake = Faker(locale, use_weighting=False)
        fake.first_name()
        mock_choices_fn.assert_called()  # select provider
        mock_choice_fn.assert_called()  # select within provider

    @patch("random.Random.choice")
    @patch("random.Random.choices", wraps=random.Random().choices)
    def test_weighting_disabled_multiple_choices(self, mock_choices_fn, mock_choice_fn):
        fake = Faker(use_weighting=False)
        fake.uri_path(deep=3)

        assert mock_choices_fn.mock_calls[0][2]["k"] == 3
        assert mock_choices_fn.mock_calls[0][2]["weights"] is None
        mock_choice_fn.assert_not_called()

    @patch("random.Random.choice")
    @patch("random.Random.choices", wraps=random.Random().choices)
    def test_weighting_enabled_multiple_choices(self, mock_choices_fn, mock_choice_fn):
        fake = Faker(use_weighting=True)
        fake.uri_path(deep=3)

        assert mock_choices_fn.mock_calls[0][2]["k"] == 3
        assert mock_choices_fn.mock_calls[0][2]["weights"] is None
        mock_choice_fn.assert_not_called()

    def test_dir_include_all_providers_attribute_in_list(self):
        fake = Faker(["en_US", "en_PH"])
        expected = set(
            dir(Faker)
            + [
                "_factories",
                "_locales",
                "_factory_map",
                "_weights",
                "_unique_proxy",
                "_optional_proxy",
            ]
        )
        for factory in fake.factories:
            expected |= {attr for attr in dir(factory) if not attr.startswith("_")}
        expected = sorted(expected)
        attributes = dir(fake)
        assert attributes == expected

    def test_copy(self):
        fake = Faker("it_IT")
        fake2 = copy.deepcopy(fake)
        assert fake.locales == fake2.locales
        assert fake.locales is not fake2.locales

    def test_pickle(self):
        fake = Faker()
        pickled = pickle.dumps(fake)
        pickle.loads(pickled)



import os
import platform
import random
import re
import sys
import time
import unittest

from datetime import date, datetime
from datetime import time as datetime_time
from datetime import timedelta, tzinfo

import freezegun
import pytest

from faker import Faker
from faker.providers.date_time import Provider as DatetimeProvider
from faker.providers.date_time import change_year
from faker.providers.date_time.ar_AA import Provider as ArProvider
from faker.providers.date_time.ar_EG import Provider as EgProvider
from faker.providers.date_time.az_AZ import Provider as AzAzProvider
from faker.providers.date_time.bn_BD import Provider as BnBdProvider
from faker.providers.date_time.cs_CZ import Provider as CsCzProvider
from faker.providers.date_time.de_AT import Provider as DeAtProvider
from faker.providers.date_time.de_DE import Provider as DeDeProvider
from faker.providers.date_time.el_GR import Provider as ElGrProvider
from faker.providers.date_time.es_ES import Provider as EsEsProvider
from faker.providers.date_time.fr_FR import Provider as FrFrProvider
from faker.providers.date_time.hy_AM import Provider as HyAmProvider
from faker.providers.date_time.it_IT import Provider as ItItProvider
from faker.providers.date_time.ja_JP import Provider as JaJpProvider
from faker.providers.date_time.nl_NL import Provider as NlProvider
from faker.providers.date_time.no_NO import Provider as NoNoProvider
from faker.providers.date_time.pl_PL import Provider as PlProvider
from faker.providers.date_time.pt_BR import Provider as PtBrProvider
from faker.providers.date_time.pt_PT import Provider as PtPtProvider
from faker.providers.date_time.ro_RO import Provider as RoRoProvider
from faker.providers.date_time.ru_RU import Provider as RuProvider
from faker.providers.date_time.sk_SK import Provider as SkSkProvider
from faker.providers.date_time.sl_SI import Provider as SlSiProvider
from faker.providers.date_time.ta_IN import Provider as TaInProvider
from faker.providers.date_time.tr_TR import Provider as TrTrProvider
from faker.providers.date_time.zh_CN import Provider as ZhCnProvider
from faker.providers.date_time.zh_TW import Provider as ZhTwProvider


def is64bit():
    return sys.maxsize > 2**32


class UTC(tzinfo):
    """
    UTC implementation taken from Python's docs.
    """

    def __repr__(self):
        return "<UTC>"

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


utc = UTC()


class TestKoKR(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("ko_KR")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert isinstance(day, str)

    def test_month(self):
        month = self.fake.month()
        assert isinstance(month, str)


class TestDateTime(unittest.TestCase):
    def setUp(self):
        self.fake = Faker()
        Faker.seed(0)

    def assertBetween(self, date, start_date, end_date):
        assert date <= end_date
        assert date >= start_date

    def test_date(self):
        date_format = "%Y-%m-%d"
        date_string = self.fake.date(pattern=date_format)
        assert isinstance(date_string, str)
        assert isinstance(datetime.strptime(date_string, date_format), datetime)

    def test_day(self):
        day = self.fake.day_of_week()
        assert isinstance(day, str)

    def test_month(self):
        month = self.fake.month()
        assert isinstance(month, str)

    def test_past_datetime(self):
        past_datetime = self.fake.past_datetime()
        assert past_datetime < datetime.now()

    def test_past_date(self):
        past_date = self.fake.past_date()
        assert past_date < date.today()

    def test_future_datetime(self):
        future_datetime, now = self.fake.future_datetime(), datetime.now()
        assert future_datetime > now

    def test_future_date(self):
        future_date = self.fake.future_date()
        assert future_date > date.today()

    def test_parse_date_time(self):
        timestamp = DatetimeProvider._parse_date_time("+30d")
        now = DatetimeProvider._parse_date_time("now")
        assert timestamp > now
        delta = timedelta(days=30)
        from_delta = DatetimeProvider._parse_date_time(delta)
        from_int = DatetimeProvider._parse_date_time(timestamp)

        assert datetime.fromtimestamp(from_delta).date() == (datetime.fromtimestamp(timestamp).date())

        assert datetime.fromtimestamp(from_int).date() == (datetime.fromtimestamp(timestamp).date())

    def test_parse_date(self):
        parsed = DatetimeProvider._parse_date("+30d")
        now = DatetimeProvider._parse_date("now")
        today = DatetimeProvider._parse_date("today")
        assert isinstance(parsed, date)
        assert isinstance(now, date)
        assert isinstance(today, date)
        assert today == date.today()
        assert now == today
        assert parsed == today + timedelta(days=30)
        assert DatetimeProvider._parse_date(datetime.now()) == today
        assert DatetimeProvider._parse_date(parsed) == parsed
        assert DatetimeProvider._parse_date(30) == parsed
        assert DatetimeProvider._parse_date(timedelta(days=30)) == parsed

    def test_timezone_conversion(self):
        from faker.providers.date_time import datetime_to_timestamp

        now = datetime.now(utc).replace(microsecond=0)
        timestamp = datetime_to_timestamp(now)
        now_back = datetime.fromtimestamp(timestamp, utc)
        assert now == now_back

        today = date.today()
        timestamp = datetime_to_timestamp(today)
        today_back = datetime.fromtimestamp(timestamp, utc).date()
        assert today == today_back

    def test_pytimezone(self):
        import dateutil

        pytz = self.fake.pytimezone()
        assert isinstance(pytz, dateutil.tz.tz.tzfile)

    def test_pytimezone_usable(self):
        pytz = self.fake.pytimezone()
        date = datetime(2000, 1, 1, tzinfo=pytz)
        assert date.tzinfo == pytz

    def test_datetimes_with_and_without_tzinfo(self):
        assert self.fake.date_time().tzinfo is None
        assert self.fake.date_time(utc).tzinfo == utc

        assert self.fake.date_time_ad().tzinfo is None
        assert self.fake.date_time_ad(utc).tzinfo == utc

        assert not self.fake.iso8601().endswith("+00:00")
        assert self.fake.iso8601(utc).endswith("+00:00")
        assert self.fake.iso8601()[10] == "T"
        assert len(self.fake.iso8601(timespec="hours")) == 13
        assert len(self.fake.iso8601(timespec="minutes")) == 16
        assert len(self.fake.iso8601(timespec="seconds")) == 19
        assert len(self.fake.iso8601(timespec="milliseconds")) == 23
        assert len(self.fake.iso8601(timespec="microseconds")) == 26
        # frequently used RFC 3339 separators
        assert self.fake.iso8601(tzinfo=utc, sep="t")[10] == "t"
        assert self.fake.iso8601(tzinfo=utc, sep=" ")[10] == " "
        assert self.fake.iso8601(tzinfo=utc, sep="_")[10] == "_"

    @pytest.mark.skipif(
        not sys.platform.startswith("win"),
        reason="windows does not support sub second precision",
    )
    def test_iso8601_fractional_seconds_win(self):
        assert len(self.fake.iso8601()) == 19

    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="non windows does support sub second precision",
    )
    def test_iso8601_fractional_seconds_non_win(self):
        assert len(self.fake.iso8601()) == 26

    def test_date_object(self):
        assert isinstance(self.fake.date_object(), date)

    def test_time_object(self):
        assert isinstance(self.fake.time_object(), datetime_time)

    def test_timedelta(self):
        delta = self.fake.time_delta(end_datetime=timedelta(seconds=60))
        assert delta.seconds <= 60

        delta = self.fake.time_delta(end_datetime=timedelta(seconds=-60))
        assert delta.seconds >= -60

        delta = self.fake.time_delta(end_datetime="+60s")
        assert delta.seconds <= 60

        delta = self.fake.time_delta(end_datetime="-60s")
        assert delta.seconds >= 60

        delta = self.fake.time_delta(end_datetime="now")
        assert delta.seconds <= 0

    def test_date_time_between_dates(self):
        timestamp_start = random.randint(0, 2000000000)
        timestamp_end = timestamp_start + 1

        datetime_start = datetime.fromtimestamp(timestamp_start)
        datetime_end = datetime.fromtimestamp(timestamp_end)

        random_date = self.fake.date_time_between_dates(datetime_start, datetime_end)
        assert datetime_start <= random_date
        assert datetime_end >= random_date

    def test_date_time_between_dates_with_no_date_overlap(self):
        with pytest.raises(ValueError):
            self.fake.date_time_between_dates("-1y", "-2y")

    def test_date_time_between_dates_with_tzinfo(self):
        timestamp_start = random.randint(0, 2000000000)
        timestamp_end = timestamp_start + 1

        datetime_start = datetime.fromtimestamp(timestamp_start, utc)
        datetime_end = datetime.fromtimestamp(timestamp_end, utc)

        random_date_naive = self.fake.date_time_between_dates(datetime_start, datetime_end)
        with pytest.raises(TypeError):
            datetime_start <= random_date_naive

        random_date = self.fake.date_time_between_dates(datetime_start, datetime_end, utc)
        assert datetime_start <= random_date
        assert datetime_end >= random_date

    def test_past_datetime_within_second(self):
        # Should not raise a ``ValueError``
        self.fake.past_datetime(start_date="+1s")

    def test_date_between_dates(self):
        date_end = date.today()
        date_start = date_end - timedelta(days=10)

        random_date = self.fake.date_between_dates(date_start, date_end)
        assert date_start <= random_date
        assert date_end >= random_date

    def test_date_time_between_long_past_dates(self):
        random_date = self.fake.date_between("-100y", "-50y")
        assert random_date

    def _datetime_to_time(self, value):
        return int(time.mktime(value.timetuple()))

    @unittest.skipUnless(is64bit(), "requires 64bit")
    def test_date_time_this_period(self):
        # test century
        this_century_start = self._datetime_to_time(
            datetime(datetime.now().year - (datetime.now().year % 100), 1, 1),
        )

        assert self._datetime_to_time(self.fake.date_time_this_century(after_now=False)) <= self._datetime_to_time(
            datetime.now()
        )
        assert self._datetime_to_time(
            self.fake.date_time_this_century(before_now=False, after_now=True)
        ) >= self._datetime_to_time(datetime.now())
        assert (
            self._datetime_to_time(self.fake.date_time_this_century(before_now=True, after_now=True))
            >= this_century_start
        )

        # test decade
        this_decade_start = self._datetime_to_time(
            datetime(datetime.now().year - (datetime.now().year % 10), 1, 1),
        )

        assert self._datetime_to_time(self.fake.date_time_this_decade(after_now=False)) <= self._datetime_to_time(
            datetime.now()
        )
        assert self._datetime_to_time(
            self.fake.date_time_this_decade(before_now=False, after_now=True)
        ) >= self._datetime_to_time(datetime.now())
        assert self._datetime_to_time(
            self.fake.date_time_this_decade(before_now=False, after_now=False)
        ) == self._datetime_to_time(datetime.now())
        assert (
            self._datetime_to_time(self.fake.date_time_this_decade(before_now=True, after_now=True))
            >= this_decade_start
        )
        # test year
        assert self._datetime_to_time(self.fake.date_time_this_year(after_now=False)) <= self._datetime_to_time(
            datetime.now()
        )
        assert self._datetime_to_time(
            self.fake.date_time_this_year(before_now=False, after_now=True)
        ) >= self._datetime_to_time(datetime.now())
        assert self._datetime_to_time(
            self.fake.date_time_this_year(before_now=False, after_now=False)
        ) == self._datetime_to_time(datetime.now())
        # test month
        assert self._datetime_to_time(self.fake.date_time_this_month(after_now=False)) <= self._datetime_to_time(
            datetime.now()
        )
        assert self._datetime_to_time(
            self.fake.date_time_this_month(before_now=False, after_now=True)
        ) >= self._datetime_to_time(datetime.now())
        assert self._datetime_to_time(
            self.fake.date_time_this_month(before_now=False, after_now=False)
        ) == self._datetime_to_time(datetime.now())

    @unittest.skipUnless(is64bit(), "requires 64bit")
    def test_date_time_this_period_with_tzinfo(self):
        # ensure all methods provide timezone aware datetimes
        with pytest.raises(TypeError):
            self.fake.date_time_this_century(before_now=False, after_now=True, tzinfo=utc) >= datetime.now()
        with pytest.raises(TypeError):
            self.fake.date_time_this_decade(after_now=False, tzinfo=utc) <= datetime.now()
        with pytest.raises(TypeError):
            self.fake.date_time_this_year(after_now=False, tzinfo=utc) <= datetime.now()
        with pytest.raises(TypeError):
            self.fake.date_time_this_month(after_now=False, tzinfo=utc) <= datetime.now()

        # test century
        assert self.fake.date_time_this_century(after_now=False, tzinfo=utc) <= datetime.now(utc)
        assert self.fake.date_time_this_century(before_now=False, after_now=True, tzinfo=utc) >= datetime.now(utc)
        assert self.fake.date_time_this_century(before_now=False, after_now=False, tzinfo=utc).replace(
            second=0, microsecond=0
        ) == datetime.now(utc).replace(second=0, microsecond=0)
        # test decade
        assert self.fake.date_time_this_decade(after_now=False, tzinfo=utc) <= datetime.now(utc)
        assert self.fake.date_time_this_decade(before_now=False, after_now=True, tzinfo=utc) >= datetime.now(utc)

        assert self.fake.date_time_this_decade(before_now=False, after_now=False, tzinfo=utc).replace(
            second=0, microsecond=0
        ) == datetime.now(utc).replace(second=0, microsecond=0)
        # test year
        assert self.fake.date_time_this_year(after_now=False, tzinfo=utc) <= datetime.now(utc)
        assert self.fake.date_time_this_year(before_now=False, after_now=True, tzinfo=utc) >= datetime.now(utc)
        assert self.fake.date_time_this_year(before_now=False, after_now=False, tzinfo=utc).replace(
            second=0, microsecond=0
        ) == datetime.now(utc).replace(second=0, microsecond=0)
        assert self.fake.date_time_this_year(before_now=True, after_now=True, tzinfo=utc).year == datetime.now(utc).year
        # test month
        assert self.fake.date_time_this_month(after_now=False, tzinfo=utc) <= datetime.now(utc)
        assert self.fake.date_time_this_month(before_now=False, after_now=True, tzinfo=utc) >= datetime.now(utc)
        assert self.fake.date_time_this_month(before_now=False, after_now=False, tzinfo=utc).replace(
            second=0, microsecond=0
        ) == datetime.now(utc).replace(second=0, microsecond=0)
        assert (
            self.fake.date_time_this_month(before_now=True, after_now=True, tzinfo=utc).month == datetime.now(utc).month
        ) and (
            self.fake.date_time_this_month(before_now=True, after_now=True, tzinfo=utc).year == datetime.now(utc).year
        )

    @unittest.skipUnless(is64bit(), "requires 64bit")
    def test_date_this_period(self):
        # test century
        assert (
            self.fake.date_this_century(before_today=True, after_today=True).strftime("%G")[:2]
            == datetime.now().strftime("%G")[:2]
        )
        assert self.fake.date_this_century(after_today=False) <= date.today()
        assert self.fake.date_this_century(before_today=False, after_today=True) >= date.today()
        assert self.fake.date_this_century(before_today=False, after_today=False) == date.today()
        # test decade
        assert (
            self.fake.date_this_decade(before_today=True, after_today=True).strftime("%G")[:3]
            == datetime.now().strftime("%G")[:3]
        )
        assert self.fake.date_this_decade(after_today=False) <= date.today()
        assert self.fake.date_this_decade(before_today=False, after_today=True) >= date.today()
        assert self.fake.date_this_decade(before_today=False, after_today=False) == date.today()
        # test year
        assert self.fake.date_this_year(before_today=True, after_today=True).year == datetime.now().year
        assert self.fake.date_this_year(after_today=False) <= date.today()
        assert self.fake.date_this_year(before_today=False, after_today=True) >= date.today()
        assert self.fake.date_this_year(before_today=False, after_today=False) == date.today()
        # test month
        assert (self.fake.date_this_month(before_today=True, after_today=True).month == datetime.now().month) and (
            self.fake.date_this_month(before_today=True, after_today=True).year == datetime.now().year
        )
        assert self.fake.date_this_month(after_today=False) <= date.today()
        assert self.fake.date_this_month(before_today=False, after_today=True) >= date.today()
        assert self.fake.date_this_month(before_today=False, after_today=False) == date.today()

    def test_date_time_between(self):
        now = datetime.now()
        _30_years_ago = change_year(now, -30)
        _20_years_ago = change_year(now, -20)

        random_datetime = self.fake.date_time_between(start_date="-30y", end_date="-20y")
        assert isinstance(random_datetime, datetime)
        self.assertBetween(random_datetime, _30_years_ago, _20_years_ago)

        now = datetime.now(tz=utc)
        _30_years_ago = change_year(now, -30)
        _20_years_ago = change_year(now, -20)

        random_datetime = self.fake.date_time_between(start_date="-30y", end_date="-20y", tzinfo=utc)
        assert isinstance(random_datetime, datetime)
        self.assertBetween(random_datetime, _30_years_ago, _20_years_ago)

    def test_date_between(self):
        today = date.today()
        _30_years_ago = change_year(today, -30)
        _20_years_ago = change_year(today, -20)

        random_date = self.fake.date_between(start_date="-30y", end_date="-20y")

        assert isinstance(random_date, date)
        self.assertBetween(random_date, _30_years_ago, _20_years_ago)

    def test_date_between_months(self):
        today = date.today()
        _2_months_ago = today - timedelta(days=2 * (365.24 / 12))
        _9_months_ago = today - timedelta(days=9 * (365.24 / 12))

        random_date = self.fake.date_between(start_date="-9M", end_date="-2M")

        assert isinstance(random_date, date)
        self.assertBetween(random_date, _9_months_ago, _2_months_ago)

    def test_parse_timedelta(self):
        from faker.providers.date_time import Provider

        td = timedelta(days=7)
        seconds = Provider._parse_timedelta(td)
        assert seconds == 604800

        seconds = Provider._parse_timedelta("+1w")
        assert seconds == 604800

        seconds = Provider._parse_timedelta("+1y")
        assert seconds == 31556736

        with pytest.raises(ValueError):
            Provider._parse_timedelta("foobar")

    def test_time_series(self):
        series = list(self.fake.time_series())
        assert len(series), 30
        assert series[1][0] - series[0][0], timedelta(days=1)

        uniform = lambda dt: random.uniform(0, 5)  # noqa
        series = list(self.fake.time_series("now", "+1w", "+1d", uniform))
        assert len(series), 7
        assert series[1][0] - series[0][0], timedelta(days=1)

        end = datetime.now() + timedelta(days=7)
        series = list(self.fake.time_series("now", end, "+1d", uniform))
        assert len(series), 7
        assert series[1][0] - series[0][0], timedelta(days=1)

        assert series[-1][0] <= end

        with pytest.raises(ValueError):
            list(self.fake.time_series("+1w", "now", "+1d", uniform))

        with pytest.raises(ValueError):
            list(self.fake.time_series("now", "+1w", "+1d", "uniform"))

        series = list(self.fake.time_series("now", end, "+1d", uniform, tzinfo=utc))
        assert len(series), 7
        assert series[1][0] - series[0][0], timedelta(days=1)

        # avoid microseconds as provider's internal parsing uses POSIX timestamps which only have second granularity
        end = datetime.now(utc).replace(microsecond=0)
        start = end - timedelta(days=15)

        series = list(self.fake.time_series(start_date=start, end_date=end, tzinfo=start.tzinfo))
        assert series[0][0] == start

    def test_unix_time(self):
        from faker.providers.date_time import datetime_to_timestamp

        for _ in range(100):
            now = datetime.now().replace(microsecond=0)
            epoch_start = datetime(1970, 1, 1, tzinfo=utc)

            # Ensure doubly-constrained unix_times are generated correctly
            start_datetime = datetime(2001, 1, 1, tzinfo=utc)
            end_datetime = datetime(2001, 1, 2, tzinfo=utc)

            constrained_unix_time = self.fake.unix_time(end_datetime=end_datetime, start_datetime=start_datetime)

            self.assertIsInstance(constrained_unix_time, (int, float))
            self.assertBetween(
                constrained_unix_time,
                datetime_to_timestamp(start_datetime),
                datetime_to_timestamp(end_datetime),
            )

            # Ensure relative unix_times partially-constrained by a start time are generated correctly
            one_day_ago = datetime.today() - timedelta(days=1)

            recent_unix_time = self.fake.unix_time(start_datetime=one_day_ago)

            self.assertIsInstance(recent_unix_time, (int, float))
            self.assertBetween(
                recent_unix_time,
                datetime_to_timestamp(one_day_ago),
                datetime_to_timestamp(now),
            )

            # Ensure relative unix_times partially-constrained by an end time are generated correctly
            one_day_after_epoch_start = datetime(1970, 1, 2, tzinfo=utc)

            distant_unix_time = self.fake.unix_time(end_datetime=one_day_after_epoch_start)

            self.assertIsInstance(distant_unix_time, (int, float))
            self.assertBetween(
                distant_unix_time,
                datetime_to_timestamp(epoch_start),
                datetime_to_timestamp(one_day_after_epoch_start),
            )

            # Ensure wide-open unix_times are generated correctly
            self.fake.unix_time()

            self.assertIsInstance(constrained_unix_time, (int, float))
            self.assertBetween(constrained_unix_time, 0, datetime_to_timestamp(now))

            # Ensure it does not throw error with startdate='now' for machines with negative offset
            if platform.system() != "Windows":
                os.environ["TZ"] = "Europe/Paris"
                time.tzset()
            self.fake.unix_time(start_datetime="now")
            if platform.system() != "Windows":
                del os.environ["TZ"]

    @pytest.mark.skipif(
        not sys.platform.startswith("win"),
        reason="windows does not support sub second precision",
    )
    def test_unix_time_win(self):
        unix_time = self.fake.unix_time()
        assert isinstance(unix_time, float)
        assert unix_time % 1 == 0.0

    @pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="non windows does support sub second precision",
    )
    def test_unix_time_non_win(self):
        unix_time = self.fake.unix_time()
        assert isinstance(unix_time, float)

    def test_change_year(self):
        _2020_06_01 = datetime.strptime("2020-06-01", "%Y-%m-%d")
        _20_years_ago = change_year(_2020_06_01, -20)
        self.assertEqual(_20_years_ago.strftime("%Y-%m-%d"), "2000-06-01")
        # Verify a leap day
        today = datetime.strptime("2020-02-29", "%Y-%m-%d")
        with self.assertRaises(ValueError):
            _11_years_ago = today.replace(year=11)
        _11_years_ago = change_year(today, -11)
        self.assertEqual(_11_years_ago.strftime("%Y-%m-%d"), "2009-03-01")
        # 0 is an invalid year, so it should still raise a ValueError
        with self.assertRaises(ValueError):
            change_year(today, -today.year)


class TestDeDe(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("de_DE")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in DeDeProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in DeDeProvider.MONTH_NAMES.values()


class TestPlPL(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("pl_PL")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in PlProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in PlProvider.MONTH_NAMES.values()


class TestHyAm(unittest.TestCase):
    """Tests date_time in the hy_AM locale"""

    def setUp(self):
        self.fake = Faker("hy_AM")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert isinstance(day, str)
        assert day in HyAmProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert isinstance(month, str)
        assert month in HyAmProvider.MONTH_NAMES.values()


class TestAr(unittest.TestCase):
    def test_ar_aa(self):
        fake = Faker("ar")
        Faker.seed(0)

        # AM/PM
        assert fake.am_pm() in ArProvider.AM_PM.values()
        # Day of week
        assert fake.century() in ArProvider.centuries
        # Month name
        assert (fake.month_name()) in (ArProvider.MONTH_NAMES.values())
        # Day of week
        assert (fake.day_of_week()) in (ArProvider.DAY_NAMES.values())

    def test_ar_eg(self):
        fake = Faker("ar_EG")
        Faker.seed(0)

        # AM/PM
        assert fake.am_pm() in ArProvider.AM_PM.values()
        # Day of week
        assert fake.century() in ArProvider.centuries
        # Day of week
        assert (fake.day_of_week()) in (ArProvider.DAY_NAMES.values())
        # Month name
        assert (fake.month_name()) in (EgProvider.MONTH_NAMES.values())
        # Month name
        assert (fake.month_name()) not in (ArProvider.MONTH_NAMES.values())


class DatesOfBirth(unittest.TestCase):
    """
    Test Dates of Birth
    """

    def setUp(self):
        self.fake = Faker()
        Faker.seed(0)

    def test_date_of_birth(self):
        dob = self.fake.date_of_birth()
        assert isinstance(dob, date)

    @freezegun.freeze_time("2020-02-29")
    def test_date_of_birth_on_leap_day(self):
        """
        Freeze the date to a leap day to verify that the date_of_birth method  does not
        raise an error
        """
        dob = self.fake.date_of_birth()
        assert isinstance(dob, date)

    def test_value_errors(self):
        with self.assertRaises(ValueError):
            self.fake.date_of_birth(minimum_age=-1)

        with self.assertRaises(ValueError):
            self.fake.date_of_birth(maximum_age=-1)

        with self.assertRaises(ValueError):
            self.fake.date_of_birth(minimum_age=-2, maximum_age=-1)

        with self.assertRaises(ValueError):
            self.fake.date_of_birth(minimum_age=5, maximum_age=4)

    def test_type_errors(self):
        with self.assertRaises(TypeError):
            self.fake.date_of_birth(minimum_age=0.5)

        with self.assertRaises(TypeError):
            self.fake.date_of_birth(maximum_age="hello")

    def test_bad_age_range(self):
        with self.assertRaises(ValueError):
            self.fake.date_of_birth(minimum_age=5, maximum_age=0)

    def test_acceptable_age_range_five_years(self):
        for _ in range(100):
            now = datetime.now(utc).date()

            days_since_now = now - now
            days_since_six_years_ago = now - change_year(now, -6)

            dob = self.fake.date_of_birth(tzinfo=utc, minimum_age=0, maximum_age=5)
            days_since_dob = now - dob

            assert isinstance(dob, date)
            assert days_since_six_years_ago > days_since_dob >= days_since_now

    def test_acceptable_age_range_eighteen_years(self):
        for _ in range(100):
            now = datetime.now(utc).date()

            days_since_now = now - now
            days_since_nineteen_years_ago = now - change_year(now, -19)

            dob = self.fake.date_of_birth(tzinfo=utc, minimum_age=0, maximum_age=18)
            days_since_dob = now - dob

            assert isinstance(dob, date)
            assert days_since_nineteen_years_ago > days_since_dob >= days_since_now

    def test_identical_age_range(self):
        for _ in range(100):
            now = datetime.now(utc).date()

            days_since_five_years_ago = now - change_year(now, -5)
            days_since_six_years_ago = now - change_year(now, -6)

            dob = self.fake.date_of_birth(tzinfo=utc, minimum_age=5, maximum_age=5)
            days_since_dob = now - dob

            assert isinstance(dob, date)
            assert days_since_six_years_ago > days_since_dob >= days_since_five_years_ago

    def test_distant_age_range(self):
        for _ in range(100):
            now = datetime.now(utc).date()

            days_since_one_hundred_years_ago = now - change_year(now, -100)
            days_since_one_hundred_eleven_years_ago = now - change_year(now, -111)

            dob = self.fake.date_of_birth(minimum_age=100, maximum_age=110)
            days_since_dob = now - dob

            assert isinstance(dob, date)
            assert days_since_one_hundred_eleven_years_ago > days_since_dob >= days_since_one_hundred_years_ago


class TestFilPh(unittest.TestCase):
    num_sample_runs = 50

    def setUp(self):
        self.setup_constants()
        self.setup_faker()

    def setup_faker(self):
        self.fake = Faker("fil_PH")
        Faker.seed(0)

    def setup_constants(self):
        from faker.providers.date_time.fil_PH import Provider

        self.day_names = Provider.DAY_NAMES.values()
        self.month_names = Provider.MONTH_NAMES.values()

    def test_PH_of_week(self):
        for _ in range(self.num_sample_runs):
            assert self.fake.day_of_week() in self.day_names

    def test_PH_month_name(self):
        for _ in range(self.num_sample_runs):
            assert self.fake.month_name() in self.month_names


class TestTlPh(TestFilPh):
    def setup_faker(self):
        self.fake = Faker("tl_PH")
        Faker.seed(0)


class TestTaIN(unittest.TestCase):
    """Tests date_time in the ta_IN locale"""

    def setUp(self):
        self.fake = Faker("ta_IN")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert isinstance(day, str)
        assert day in TaInProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert isinstance(month, str)
        assert month in TaInProvider.MONTH_NAMES.values()


class TestRuRu(unittest.TestCase):
    """Tests date_time in the ru_RU locale"""

    def setUp(self):
        self.fake = Faker("ru_RU")
        Faker.seed(0)

    def test_day(self):
        for _ in range(50):
            day = self.fake.day_of_week()
            assert isinstance(day, str)
            assert day in RuProvider.DAY_NAMES.values()

    def test_month(self):
        for _ in range(50):
            month = self.fake.month_name()
            assert isinstance(month, str)
            assert month in RuProvider.MONTH_NAMES.values()

    def test_timezone(self):
        for _ in range(100):
            timezone = self.fake.timezone()
            assert isinstance(timezone, str)
            assert re.match(r"[А-Яа-я]", timezone)


class TestAzAz(unittest.TestCase):
    """Test az_AZ date_time provider methods"""

    def setUp(self):
        self.fake = Faker("az_AZ")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in AzAzProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in AzAzProvider.MONTH_NAMES.values()


class TestCsCz(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("cs_CZ")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in CsCzProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in CsCzProvider.MONTH_NAMES.values()


class TestDeAt(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("de_AT")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in DeAtProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in DeAtProvider.MONTH_NAMES.values()


class TestEsEs(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("es_ES")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in EsEsProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in EsEsProvider.MONTH_NAMES.values()


class TestItIt(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("it_IT")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in ItItProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in ItItProvider.MONTH_NAMES.values()


class TestSkSk(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("sk_SK")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in SkSkProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in SkSkProvider.MONTH_NAMES.values()


class TestSlSi(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("sl_SI")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in SlSiProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in SlSiProvider.MONTH_NAMES.values()


class TestThTh(unittest.TestCase):
    num_sample_runs = 50

    def setUp(self):
        self.fake = Faker("th_TH")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert isinstance(day, str)
        assert day.startswith("วัน")

    def test_month(self):
        month = self.fake.month_name()
        assert isinstance(month, str)

    def test_date(self):
        # default format is "%-d %b %Y"
        # (date with no padding, abbreviated month, full year [4 digits])
        date = self.fake.date(thai_digit=True)
        for _ in range(self.num_sample_runs):
            assert re.fullmatch(
                r"[๐-๙]{1,2} "
                r"(ม\.ค\.|ก\.พ\.|มี\.ค\.|เม\.ย\.|พ\.ค\.|มิ\.ย\.|ก\.ค\.|ส\.ค\.|ก\.ย\.|ต\.ค\.|พ\.ย\.|ธ\.ค\.) "
                r"[๐-๙]{4}",
                date,
            )

    def test_time(self):
        time = self.fake.time(thai_digit=True)
        for _ in range(self.num_sample_runs):
            assert re.fullmatch(r"[๐-๙]{2}:[๐-๙]{2}:[๐-๙]{2}", time)

    def test_century(self):
        century = self.fake.century()
        assert isinstance(century, str)
        assert len(century) <= 2
        century = self.fake.century(thai_digit=True)
        for _ in range(self.num_sample_runs):
            assert re.fullmatch(r"[๑-๒]?[๐-๙]", century)

    def test_date_pattern(self):
        # unsupported or incomplete directive
        date = self.fake.date("%Q")
        assert date == "Q"
        date = self.fake.date("%%")
        assert date == "%"
        date = self.fake.date("%-")
        assert date == "-"
        date = self.fake.date("% ")
        assert date == " "
        date = self.fake.date("%0")
        assert date == "0"
        date = self.fake.date("%")
        assert date == "%"

        # may be not supported on Windows, try for coverage
        date = self.fake.date("%s")
        date = self.fake.date("%f")

        # National representation of the full weekday name
        date = self.fake.date("%A")
        assert isinstance(date, str)

        # National representation of the abbreviated weekday
        date = self.fake.date("%a")
        assert isinstance(date, str)
        assert len(date) <= 2

        # National representation of the full month name
        date = self.fake.date("%B")
        assert isinstance(date, str)
        assert "." not in date

        # National representation of the abbreviated month name
        date = self.fake.date("%b")
        assert isinstance(date, str)
        assert "." in date

        # Century as decimal number
        date = self.fake.date("%C")
        assert isinstance(date, str)

        # Locale’s appropriate date and time representation
        # พ   6 ต.ค. 01:40:00 2519  <-- left-aligned weekday, right-aligned day
        date = self.fake.date("%c")
        assert isinstance(date, str)

        # Equivalent to ``%m/%d/%y''
        date = self.fake.date("%D")
        assert isinstance(date, str)

        # Equivalent to ``%Y-%m-%d''
        date = self.fake.date("%F")
        assert isinstance(date, str)

        # ISO 8601 year with century representing the year that contains
        # the greater part of the ISO week (%V). Monday as the first day
        # of the week.
        date = self.fake.date("%G")
        assert isinstance(date, str)

        # Same year as in ``%G'',
        # but as a decimal number without century (00-99).
        date = self.fake.date("%g")
        assert isinstance(date, str)
        assert len(date) <= 2

        # BSD extension, ' 6-ต.ค.-2519'
        date = self.fake.date("%v")
        assert isinstance(date, str)

        # Locale’s appropriate time representation.
        date = self.fake.date("%X")
        assert isinstance(date, str)

        # Locale’s appropriate date representation.
        date = self.fake.date("%x")
        assert isinstance(date, str)

        # Year with century
        date = self.fake.date("%Y")
        assert isinstance(date, str)

        # Year without century
        date = self.fake.date("%y")
        assert isinstance(date, str)
        assert len(date) <= 2

        # National representation of the date and time
        # (the format is similar to that produced by date(1))
        # Wed  6 Oct 1976 01:40:00
        date = self.fake.date("%+")
        assert isinstance(date, str)

        # GNU libc extension,
        # no padding
        for _ in range(self.num_sample_runs):
            date = self.fake.date("%-d")
            assert isinstance(date, str)
            assert date[0] != "0"

        # GNU libc extension,
        # explicitly specify space (" ") for padding
        for _ in range(self.num_sample_runs):
            date = self.fake.date("%_d")
            assert isinstance(date, str)
            assert date[0] != "0"

        # GNU libc extension,
        # explicitly specify zero ("0") for padding
        for _ in range(self.num_sample_runs):
            date = self.fake.date("%0v")
            assert isinstance(date, str)
            assert date[0] != " "

        # GNU libc extension,
        # convert to upper case
        date = self.fake.date("%^p")
        assert isinstance(date, str)
        assert date.isupper()

        # GNU libc extension,
        # swap case - useful for %Z
        date = self.fake.date("%#p")
        assert isinstance(date, str)
        assert date.islower()

        # POSIX extension,
        # uses the locale's alternative representation
        # Not implemented yet
        # swap case - useful for %Z
        date = self.fake.date("%Ed")
        assert isinstance(date, str)

        # POSIX extension,
        # uses the locale's alternative numeric symbols
        date = self.fake.date("%Od")
        assert isinstance(date, str)
        assert date[0] not in "0123456789"


class TestTrTr(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("tr_TR")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in TrTrProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in TrTrProvider.MONTH_NAMES.values()


class TestPtBr(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("pt_BR")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in PtBrProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in PtBrProvider.MONTH_NAMES.values()


class TestPtPt(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("pt_PT")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in PtPtProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in PtPtProvider.MONTH_NAMES.values()


class TestRoRo(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("ro_RO")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in RoRoProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in RoRoProvider.MONTH_NAMES.values()


class TestBnBd(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("bn_BD")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in BnBdProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in BnBdProvider.MONTH_NAMES.values()


class TestNlNl(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("nl_NL")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in NlProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in NlProvider.MONTH_NAMES.values()


class TestElGr(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("el-GR")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in ElGrProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in ElGrProvider.MONTH_NAMES.values()


class TestZhCn(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("zh-CN")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in ZhCnProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in ZhCnProvider.MONTH_NAMES.values()


class TestZhTw(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("zh-TW")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in ZhTwProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in ZhTwProvider.MONTH_NAMES.values()

    def test_year(self):
        year = self.fake.year()
        assert isinstance(year, str)
        assert year.isdigit()
        assert len(year) >= 4

    def test_minguo_year(self):
        year = self.fake.minguo_year()
        assert isinstance(year, str)
        assert year.isdigit()
        assert 1 <= len(year) <= 3


class TestNoNo(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("no-NO")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in NoNoProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in NoNoProvider.MONTH_NAMES.values()


class TestFrFr(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("fr-FR")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in FrFrProvider.DAY_NAMES.values()

    def test_month(self):
        day = self.fake.month_name()
        assert day in FrFrProvider.MONTH_NAMES.values()


class TestFrCa(unittest.TestCase):
    # Both fr_FR and fr_CA share the same date format so we use the same test suite
    def setUp(self):
        self.fake = Faker("fr-CA")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in FrFrProvider.DAY_NAMES.values()

    def test_month(self):
        day = self.fake.month_name()
        assert day in FrFrProvider.MONTH_NAMES.values()


class TestJaJp(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("ja_JP")
        Faker.seed(0)

    def test_day(self):
        day = self.fake.day_of_week()
        assert day in JaJpProvider.DAY_NAMES.values()

    def test_month(self):
        month = self.fake.month_name()
        assert month in JaJpProvider.MONTH_NAMES.values()

    def test_traditional_month(self):
        month = self.fake.traditional_month_name()
        assert month in JaJpProvider.TRADITIONAL_MONTH_NAMES.values()



from collections import OrderedDict

from ..en import Provider as AddressProvider


class Provider(AddressProvider):
    city_prefixes = ("North", "East", "West", "South", "New", "Lake", "Port")
    city_suffixes = (
        "town",
        "ton",
        "land",
        "ville",
        "berg",
        "burgh",
        "borough",
        "bury",
        "view",
        "port",
        "mouth",
        "stad",
        "furt",
        "chester",
        "mouth",
        "fort",
        "haven",
        "side",
        "shire",
    )
    counties = (
        "Bedfordshire",
        "Buckinghamshire",
        "Cambridgeshire",
        "Cheshire",
        "Cleveland",
        "Cornwall",
        "Cumbria",
        "Derbyshire",
        "Devon",
        "Dorset",
        "Durham",
        "East Sussex",
        "Essex",
        "Gloucestershire",
        "Greater London",
        "Greater Manchester",
        "Hampshire",
        "Hertfordshire",
        "Kent",
        "Lancashire",
        "Leicestershire",
        "Lincolnshire",
        "Merseyside",
        "Norfolk",
        "North Yorkshire",
        "Northamptonshire",
        "Northumberland",
        "Nottinghamshire",
        "Oxfordshire",
        "Shropshire",
        "Somerset",
        "South Yorkshire",
        "Staffordshire",
        "Suffolk",
        "Surrey",
        "Tyne and Wear",
        "Warwickshire",
        "West Berkshire",
        "West Midlands",
        "West Sussex",
        "West Yorkshire",
        "Wiltshire",
        "Worcestershire",
        "Flintshire",
        "Glamorgan",
        "Merionethshire",
        "Monmouthshire",
        "Montgomeryshire",
        "Pembrokeshire",
        "Radnorshire",
        "Anglesey",
        "Breconshire",
        "Caernarvonshire",
        "Cardiganshire",
        "Carmarthenshire",
        "Denbighshire",
        "Aberdeen City",
        "Aberdeenshire",
        "Angus",
        "Argyll and Bute",
        "City of Edinburgh",
        "Clackmannanshire",
        "Dumfries and Galloway",
        "Dundee City",
        "East Ayrshire",
        "East Dunbartonshire",
        "East Lothian",
        "East Renfrewshire",
        "Eilean Siar",
        "Falkirk",
        "Fife",
        "Glasgow City",
        "Highland",
        "Inverclyde",
        "Midlothian",
        "Moray",
        "North Ayrshire",
        "North Lanarkshire",
        "Orkney Islands",
        "Perth and Kinross",
        "Renfrewshire",
        "Scottish Borders",
        "Shetland Islands",
        "South Ayrshire",
        "South Lanarkshire",
        "Stirling",
        "West Dunbartonshire",
        "West Lothian",
        "Antrim",
        "Armagh",
        "Down",
        "Fermanagh",
        "Derry and Londonderry",
        "Tyrone",
    )
    building_number_formats = ("#", "##", "###")
    street_suffixes = (
        "alley",
        "avenue",
        "branch",
        "bridge",
        "brook",
        "brooks",
        "burg",
        "burgs",
        "bypass",
        "camp",
        "canyon",
        "cape",
        "causeway",
        "center",
        "centers",
        "circle",
        "circles",
        "cliff",
        "cliffs",
        "club",
        "common",
        "corner",
        "corners",
        "course",
        "court",
        "courts",
        "cove",
        "coves",
        "creek",
        "crescent",
        "crest",
        "crossing",
        "crossroad",
        "curve",
        "dale",
        "dam",
        "divide",
        "drive",
        "drive",
        "drives",
        "estate",
        "estates",
        "expressway",
        "extension",
        "extensions",
        "fall",
        "falls",
        "ferry",
        "field",
        "fields",
        "flat",
        "flats",
        "ford",
        "fords",
        "forest",
        "forge",
        "forges",
        "fork",
        "forks",
        "fort",
        "freeway",
        "garden",
        "gardens",
        "gateway",
        "glen",
        "glens",
        "green",
        "greens",
        "grove",
        "groves",
        "harbor",
        "harbors",
        "haven",
        "heights",
        "highway",
        "hill",
        "hills",
        "hollow",
        "inlet",
        "inlet",
        "island",
        "island",
        "islands",
        "islands",
        "isle",
        "isle",
        "junction",
        "junctions",
        "key",
        "keys",
        "knoll",
        "knolls",
        "lake",
        "lakes",
        "land",
        "landing",
        "lane",
        "light",
        "lights",
        "loaf",
        "lock",
        "locks",
        "locks",
        "lodge",
        "lodge",
        "loop",
        "mall",
        "manor",
        "manors",
        "meadow",
        "meadows",
        "mews",
        "mill",
        "mills",
        "mission",
        "mission",
        "motorway",
        "mount",
        "mountain",
        "mountain",
        "mountains",
        "mountains",
        "neck",
        "orchard",
        "oval",
        "overpass",
        "park",
        "parks",
        "parkway",
        "parkways",
        "pass",
        "passage",
        "path",
        "pike",
        "pine",
        "pines",
        "place",
        "plain",
        "plains",
        "plains",
        "plaza",
        "plaza",
        "point",
        "points",
        "port",
        "port",
        "ports",
        "ports",
        "prairie",
        "prairie",
        "radial",
        "ramp",
        "ranch",
        "rapid",
        "rapids",
        "rest",
        "ridge",
        "ridges",
        "river",
        "road",
        "road",
        "roads",
        "roads",
        "route",
        "row",
        "rue",
        "run",
        "shoal",
        "shoals",
        "shore",
        "shores",
        "skyway",
        "spring",
        "springs",
        "springs",
        "spur",
        "spurs",
        "square",
        "square",
        "squares",
        "squares",
        "station",
        "station",
        "stravenue",
        "stravenue",
        "stream",
        "stream",
        "street",
        "street",
        "streets",
        "summit",
        "summit",
        "terrace",
        "throughway",
        "trace",
        "track",
        "trafficway",
        "trail",
        "trail",
        "tunnel",
        "tunnel",
        "turnpike",
        "turnpike",
        "underpass",
        "union",
        "unions",
        "valley",
        "valleys",
        "via",
        "viaduct",
        "view",
        "views",
        "village",
        "village",
        "villages",
        "ville",
        "vista",
        "vista",
        "walk",
        "walks",
        "wall",
        "way",
        "ways",
        "well",
        "wells",
    )

    POSTAL_ZONES = (
        "AB",
        "AL",
        "B",
        "BA",
        "BB",
        "BD",
        "BH",
        "BL",
        "BN",
        "BR",
        "BS",
        "BT",
        "CA",
        "CB",
        "CF",
        "CH",
        "CM",
        "CO",
        "CR",
        "CT",
        "CV",
        "CW",
        "DA",
        "DD",
        "DE",
        "DG",
        "DH",
        "DL",
        "DN",
        "DT",
        "DY",
        "E",
        "EC",
        "EH",
        "EN",
        "EX",
        "FK",
        "FY",
        "G",
        "GL",
        "GY",
        "GU",
        "HA",
        "HD",
        "HG",
        "HP",
        "HR",
        "HS",
        "HU",
        "HX",
        "IG",
        "IM",
        "IP",
        "IV",
        "JE",
        "KA",
        "KT",
        "KW",
        "KY",
        "L",
        "LA",
        "LD",
        "LE",
        "LL",
        "LN",
        "LS",
        "LU",
        "M",
        "ME",
        "MK",
        "ML",
        "N",
        "NE",
        "NG",
        "NN",
        "NP",
        "NR",
        "NW",
        "OL",
        "OX",
        "PA",
        "PE",
        "PH",
        "PL",
        "PO",
        "PR",
        "RG",
        "RH",
        "RM",
        "S",
        "SA",
        "SE",
        "SG",
        "SK",
        "SL",
        "SM",
        "SN",
        "SO",
        "SP",
        "SR",
        "SS",
        "ST",
        "SW",
        "SY",
        "TA",
        "TD",
        "TF",
        "TN",
        "TQ",
        "TR",
        "TS",
        "TW",
        "UB",
        "W",
        "WA",
        "WC",
        "WD",
        "WF",
        "WN",
        "WR",
        "WS",
        "WV",
        "YO",
        "ZE",
    )

    POSTAL_ZONES_ONE_CHAR = [zone for zone in POSTAL_ZONES if len(zone) == 1]
    POSTAL_ZONES_TWO_CHARS = [zone for zone in POSTAL_ZONES if len(zone) == 2]

    postcode_formats = (
        "AN NEE",
        "ANN NEE",
        "PN NEE",
        "PNN NEE",
        "ANC NEE",
        "PND NEE",
    )

    _postcode_sets = OrderedDict(
        (
            (" ", " "),
            ("N", [str(i) for i in range(0, 10)]),
            ("A", POSTAL_ZONES_ONE_CHAR),
            ("B", "ABCDEFGHKLMNOPQRSTUVWXY"),
            ("C", "ABCDEFGHJKSTUW"),
            ("D", "ABEHMNPRVWXY"),
            ("E", "ABDEFGHJLNPQRSTUWXYZ"),
            ("P", POSTAL_ZONES_TWO_CHARS),
        )
    )

    city_formats = (
        "{{city_prefix}} {{first_name}}{{city_suffix}}",
        "{{city_prefix}} {{first_name}}",
        "{{first_name}}{{city_suffix}}",
        "{{last_name}}{{city_suffix}}",
    )
    street_name_formats = (
        "{{first_name}} {{street_suffix}}",
        "{{last_name}} {{street_suffix}}",
    )
    street_address_formats = (
        "{{building_number}} {{street_name}}",
        "{{secondary_address}}\n{{street_name}}",
    )
    address_formats = ("{{street_address}}\n{{city}}\n{{postcode}}",)
    secondary_address_formats = (
        "Flat #",
        "Flat ##",
        "Flat ##?",
        "Studio #",
        "Studio ##",
        "Studio ##?",
    )

    def postcode(self) -> str:
        """
        See
        http://web.archive.org/web/20090930140939/http://www.govtalk.gov.uk/gdsc/html/noframes/PostCode-2-1-Release.htm
        """
        postcode = ""
        pattern: str = self.random_element(self.postcode_formats)
        for placeholder in pattern:
            postcode += self.random_element(self._postcode_sets[placeholder])
        return postcode

    def city_prefix(self) -> str:
        return self.random_element(self.city_prefixes)

    def secondary_address(self) -> str:
        return self.bothify(self.random_element(self.secondary_address_formats))

    def administrative_unit(self) -> str:
        return self.random_element(self.counties)

    county = administrative_unit



from .. import Provider as InternetProvider


class Provider(InternetProvider):
    free_email_domains = (
        "gmail.com",
        "hotmail.com",
        "yahoo.com",
        "net.hr",
        "zg.t-com.hr",
        "inet.hr",
        "t.ht.hr",
        "vip.hr",
        "globalnet.hr",
        "xnet.hr",
        "yahoo.hr",
        "zagreb.hr",
    )

    tlds = ("hr", "com", "com.hr", "info", "org", "net", "biz")

    replacements = (
        ("č", "c"),
        ("Č", "C"),
        ("ć", "c"),
        ("Ć", "C"),
        ("đ", "dj"),
        ("Đ", "Dj"),
        ("š", "s"),
        ("Š", "S"),
        ("ž", "z"),
        ("Ž", "Z"),
    )



from .codes import codes


def unidecode(txt: str) -> str:
    chars = ""
    for ch in txt:
        codepoint = ord(ch)

        try:
            chars += codes[codepoint]
        except IndexError:
            pass
    return chars
